package be.iminds.ilabt.jfed.rspec.statistics;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerGlimpse;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerGlimpseBuilder;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RspecSource;
import be.iminds.ilabt.jfed.util.common.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Timestamp;
import java.util.*;
import java.util.stream.Collectors;

/**
 * RspecStatisticsMaker
 */
public class RspecStatisticsMaker {
    private static final Logger ownLOG = LoggerFactory.getLogger(RspecStatisticsMaker.class);
    private final org.slf4j.Logger LOG;

    public RspecStatisticsMaker() {
        this(ownLOG);
    }
    public RspecStatisticsMaker(Logger LOG) {
        this.LOG = LOG;
    }
    public static Logger getOwnLOG() {
        return ownLOG;
    }

    /**
     * Create statistics on number of nodes
     * @param modelRspec
     * @return statistics in 2 versions
     */
    public Pair<RSpecStatistics, ServerGlimpseBuilder> make(@Nonnull ModelRspec modelRspec,
                                                            boolean isAvailableAdvertisement,
                                                            @Nonnull Server targetServer,
                                                            @Nullable Timestamp timeStamp) {
        //these 3 are used to report certain events only the first time
        boolean debugReportedNodeWithoutLocation = false;
        boolean debugReportedNodeWithLocation = false;
        boolean debugReportedNodeWithLocationError = false;

        RSpecStatisticsBuilder builder = new RSpecStatisticsBuilder();
        ServerGlimpseBuilder glimpse = new ServerGlimpseBuilder();
        glimpse.setServer(targetServer);
        glimpse.setAdvertisementTimestamp(timeStamp);

        boolean debugHasAvailable = false;
        boolean debugExclusive = false;
        boolean debugExclusiveTrue = false;
        boolean debugExclusiveFalse = false;

        //all counts are zero initially

        for (RspecNode n : modelRspec.getNodes()) {
            try {
                if (isAvailableAdvertisement && Objects.equals(n.getAvailable(), Boolean.FALSE)) {
                    LOG.warn("We requested only available nodes, but got a node that was not available. The node will be ignored. component_id=" + n.getComponentId());
                    builder.incCountTotal();
                    continue;
                }

                if (n.getComponentManagerId() != null) {
                    glimpse.addComponentManagerId(n.getComponentManagerId().getValue());
                }

                if (n.getExclusive() != null) {
                    debugExclusive = true;
                    if (n.getExclusive())
                        debugExclusiveTrue = true;
                    else
                        debugExclusiveFalse = true;
                }
                if (n.getAvailable() != null) {
                    debugHasAvailable = true;
                }

                boolean doCount = true;
                boolean doCountTotal = true;
                Integer curCount = null;

                boolean doCountRaw = true;
                boolean doCountRawTotal = true;
                boolean doCountVm = true;
                boolean doCountVmTotal = true;

                int curCountRaw = 1;
                int curCountRawTotal = 1;
                int curCountVm = 0;
                int curCountVmTotal = 0;

                if (/*!isAvailableAdvertisement &&*/ Objects.equals(n.getAvailable(), Boolean.FALSE)) {
                    doCount = false;
                    doCountVm = false;
                    doCountRaw = false;
                }

                Set<String> hwTypeNames = new HashSet<>();
                if (n.getHardwareTypes() != null) {
                    for (HardwareType hwType : n.getHardwareTypes()) {
                        String hwTypeName = hwType.getName();
                        if (hwTypeName != null)
                            hwTypeNames.add(hwTypeName);
                        if (hwTypeName != null &&
                                mustIgnoreNode(n, hwTypeName, targetServer)) {
                            doCount = false;
                            doCountTotal = false;
                            doCountVm = false;
                            doCountRaw = false;
                            doCountVmTotal = false;
                            doCountRawTotal = false;
                            builder.addExplicitSkippedHwType(hwTypeName);
                        }
                        if (isVm(n, hwTypeName, targetServer)) {
                            doCountRaw = false;
                            doCountRawTotal = false;

                            //emulab specific (but also used by other testbeds)
                            if (hwType.hasTypeSlots() && hwType.getTypeSlotsInt() != null &&
                                    mustCountVmTypeSlots(n, hwTypeName, targetServer)) {
                                if (curCount != null && hwType.getTypeSlotsInt() != ((int) curCount)) {
                                    ownLOG.warn("Count problem: Already saw a different count for hardware_type with type slots. Will overwrite " + curCount + " with " + hwType.getTypeSlotsInt());
                                    LOG.warn("Count problem: Already saw a different count for hardware_type with type slots. Will overwrite " + curCount + " with " + hwType.getTypeSlotsInt());
                                }
                                curCount = hwType.getTypeSlotsInt();
                                curCountVm = hwType.getTypeSlotsInt();

                                if (curCount < 0 || curCountVm < 0) {
                                    LOG.warn("Count problem: node "+n.getComponentId()+" has typeslot with negative value: "+curCount+"  (will consider this 0!)");
                                    curCount = 0;
                                    curCountVm = 0;
                                    builder.addCountWarning("Count problem: node "+n.getComponentId()+" has typeslot with negative value: "+curCount+"  (will consider this 0!)");
                                }

                                curCountVmTotal = getTotalVmCountPerNode(curCountVm, n, hwTypeName, targetServer);

                                if (includeVmCountInTotal(n, hwTypeName, targetServer)) {
                                    doCount = true;
                                    doCountTotal = true;
                                } else {
                                    doCount = false;
                                    doCountTotal = false;
                                }
                            }
                        }
                        if (Objects.equals(n.getExclusive(), Boolean.TRUE)) {
                            doCountVm = false;
                            doCountVmTotal = false;
                        }
                    }
                }
                final Server tServ = targetServer;
                Set<String> hwTypeNamesForServerGlimpse =
                        hwTypeNames.stream()
                                .filter(Objects::nonNull)
                                .filter((hwTypeName) -> { return !mustExcludeHardwareTypeFromServerGlimpse(n, hwTypeName, tServ); })
                                .collect(Collectors.toSet());
                if (n.getComponentId() != null) {
                    boolean unavailable = Objects.equals(n.getAvailable(), Boolean.FALSE);

                    ServerGlimpse.AdvertisementInfo.ComponentDetails.Location glimpseLocation;
                    if (n.getLocation() != null) {
                        NodeLocation rspecLocation = n.getLocation();
                        try {
                            glimpseLocation =
                                    new ServerGlimpse.AdvertisementInfo.ComponentDetails.Location(
                                            rspecLocation.getCountry(),
                                            rspecLocation.getLatitude() == null ? null : Double.parseDouble(rspecLocation.getLatitude()),
                                            rspecLocation.getLongitude() == null ? null : Double.parseDouble(rspecLocation.getLongitude())
                                    );
                            if (LOG.isDebugEnabled()) {
                                if (!debugReportedNodeWithLocation) {
                                    LOG.debug("Saw a node with node location " +
                                            "id=\""+n.getComponentId()+"\" (this is only reported once)");
                                }
                                debugReportedNodeWithLocation = true;
                            }
                        } catch (NumberFormatException e) {
                            builder.addCountWarning("Failed to parse location from rspec node location " +
                                    "lat=\""+rspecLocation.getLatitude()+"\" " +
                                    "lng=\""+rspecLocation.getLongitude()+"\" " +
                                    "id=\""+n.getComponentId()+"\" ");
                            if (LOG.isDebugEnabled()) {
                                if (!debugReportedNodeWithLocationError) {
                                    LOG.debug("Failed to parse location from rspec node location " +
                                            "lat=\""+rspecLocation.getLatitude()+"\" " +
                                            "lng=\""+rspecLocation.getLongitude()+"\" " +
                                            "id=\""+n.getComponentId()+"\" " +
                                            "(this is only reported once)", e);
                                }
                                debugReportedNodeWithoutLocation = true;
                            }
                            glimpseLocation = null;
                        }
                    } else {
                        if (LOG.isDebugEnabled()) {
                            if (!debugReportedNodeWithoutLocation) {
                                LOG.debug("Saw a node without node location " +
                                        "id=\""+n.getComponentId()+"\" (this is only reported once)");
                            }
                            debugReportedNodeWithoutLocation = true;
                        }
                        glimpseLocation = null;
                    }

                    ServerGlimpse.AdvertisementInfo.ComponentDetails componentDetails =
                            new ServerGlimpse.AdvertisementInfo.ComponentDetails(
                                    n.getComponentId().getValue(),
                                    new ArrayList(hwTypeNamesForServerGlimpse),
                                    !unavailable,
                                    glimpseLocation);
                    glimpse.addComponentDetails(componentDetails);
                }

                Set<String> sliverTypeNames = new TreeSet<>();
                if (n.getSliverTypes() != null) {
                    for (SliverType sliverType : n.getSliverTypes()) {
                        if (sliverType.getName() != null)
                            sliverTypeNames.add(sliverType.getName());
                        String oneDefaultDiskImageId = null;
                        if (sliverType.getDiskImages() != null) {
                            for (DiskImage diskImage : sliverType.getDiskImages()) {
                                if (diskImage.isDefault() != null && diskImage.isDefault() && diskImage.getName() != null)
                                    oneDefaultDiskImageId = diskImage.getName();
                                glimpse.mergeDiskImage(diskImage.getName(), diskImage.getUrl(),
                                        diskImage.getDescription(), diskImage.getVersion(),
                                        diskImage.getOs(), sliverType.getName());
                            }
                        }
                        if (sliverType.getName() != null) {
                            glimpse.mergeSliverTypeInfo(sliverType.getName(), hwTypeNamesForServerGlimpse, oneDefaultDiskImageId);
                        }
                    }
                }
                if (sliverTypeNames.isEmpty()) {
                    glimpse.addNoSliverHardwareType(hwTypeNamesForServerGlimpse);
                }

                try {
                    LOG.debug(
                            String.format("Counting \"%s\" of hwType \"%s\" exc=%s, as raw=%d/%d vm=%d/%d",
                                    friendlyName(n),
                                    n.getHardwareTypes().stream()
                                            .map(HardwareType::getName)
                                            .collect(Collectors.joining("\" and \"")),
                                    n.getExclusive(),
                                    !doCountRaw ? 0 : curCountRaw,
                                    !doCountRawTotal ? 0 : curCountRawTotal,
                                    !doCountVm ? 0 : curCountVm,
                                    !doCountVmTotal ? 0 : curCountVmTotal));
                } catch (Exception e) {
                    LOG.error("Error while printing debug info. Will be ignored. "+e.getMessage());
                }

                if (doCountRaw) {
                    builder.incCountRaw(curCountRaw);
                    for (String hwType : hwTypeNamesForServerGlimpse) {
                        builder.incAvailableForHwType(hwType, curCountRaw);
                    }
                }
                if (doCountRawTotal) {
                    builder.incCountRawTotal(curCountRawTotal);
                    for (String hwType : hwTypeNamesForServerGlimpse) {
                        builder.incTotalForHwType(hwType, curCountRawTotal);
                    }
                }
                if (doCountVm)
                    builder.incCountVm(curCountVm);
                if (doCountVmTotal) {
                    if (curCountVm > curCountVmTotal) {
                        LOG.warn("More VMs available (" + curCountVm + ") than supposed maximum (" + curCountVmTotal + ") -> Will fix by updating max");
                        builder.addCountWarning("More VMs available (" + curCountVm + ") than supposed maximum (" + curCountVmTotal + ")");
                        curCountVmTotal = curCountVm;
                    }
                    builder.incCountVmTotal(curCountVmTotal);
                }

                if (doCount) {
                    builder.addCountedHwTypes(hwTypeNames);
                    builder.incCount(curCount == null ? 1 : curCount);
                } else {
                    builder.addImplicitSkippedHwTypes(hwTypeNames);
                    builder.incSkippedCount(); //ignore type_slots for skipped_count
                }

                if (doCountTotal) {
                    builder.incCountTotal(curCount == null ? 1 : curCount);
                }
            } catch (Throwable e) {
                ownLOG.warn("Something went wrong while trying to count a node. This error can be ignored but it might cause a bad count.", e);
                LOG.warn("Something went wrong while trying to count a node. This error can be ignored but it might cause a bad count.", e);
//                res.count = modelRspec.getNodes().size();
//                res.countTotal = modelRspec.getNodes().size();
                //do nothing, which probably ignores the node
            }
        }

        if (builder.getCount() > builder.getCountRaw() && builder.getCountRaw() > 0 && builder.getCountVm() > 0) {
            LOG.info("           Automatic summary count fixup: "+builder.getCount()+" -> " + builder.getCountRaw()+
                    " (countRaw="+builder.getCountRaw()+" countVm="+builder.getCountVm()+")");
            builder.setCount(builder.getCountRaw());
        }
        if (builder.getCount() == 0 && builder.getCountRaw() == 0 && builder.getCountVm() > 0) {
            LOG.info("           Automatic summary count fixup: 0 -> " + builder.getCountVm());
            builder.setCount(builder.getCountVm());
        }

        LOG.info("           Counted Hardware Types: " + builder.getCountedHwTypes());
        LOG.info("Implicitly Skipped Hardware Types: " + builder.getImplicitSkippedHwTypes());
        LOG.info("Explicitly Skipped Hardware Types: " + builder.getExplicitSkippedHwTypes());

        LOG.debug("Advertisement RSpec nodes have available: " + debugHasAvailable);
        LOG.debug("Advertisement RSpec nodes have exclusive: " + debugExclusive +
                (debugExclusive ? (" hasTrue=" + debugExclusiveTrue + " hasFalse=" + debugExclusiveFalse) : ""));

        LOG.info("Result count=" + builder.getCount() +
                " countTotal=" + builder.getCountTotal() +
                " skippedCount=" + builder.getSkippedCount());

        LOG.info("Result count raw " + builder.getCountRaw() + " of " + builder.getCountRawTotal());
        LOG.info("Result count  VM " + builder.getCountVm() + " of " + builder.getCountVmTotal());

        List<? extends OpenflowDataPath> openflowDataPaths = modelRspec.getOpenflowDataPaths();
        for (OpenflowDataPath dp : openflowDataPaths) {
            if (dp.getDpid() != null) {
                builder.incCount();
                builder.incCountOpenflow();
                builder.incCountOpenflowTotal();
            }
        }
        LOG.info("Result count  openflow " + builder.getCountOpenflow() + " of " + builder.getCountOpenflowTotal());

        HardwareTypeInfo hardwareTypeInfo = modelRspec.getHardwareTypeInfo();
        if (hardwareTypeInfo != null) {
            //convert HardwareTypeInfo in RSpec to HardwareTypeInfo in ServerGlimpse
            URI htmlOverview;
            try {
                htmlOverview = hardwareTypeInfo.getHtmlOverview() == null ? null :
                        new URI(hardwareTypeInfo.getHtmlOverview());
            } catch (URISyntaxException e) {
                LOG.warn("Invalid URI in hardwaretypeinfo overview: "+hardwareTypeInfo.getHtmlOverview());
                htmlOverview = null;
            }

            List<ServerGlimpse.AdvertisementInfo.HardwareTypeInfo.HardwareTypeInfoEntry> entries = new ArrayList<>();
            for (HardwareTypeInfo.HardwareTypeInfoEntry ee: hardwareTypeInfo.getHardwareTypeInfoEntries()) {
                URI eHtmlDescription;
                try {
                    eHtmlDescription = ee.getHtmlDescription() == null ? null :
                            new URI(ee.getHtmlDescription());
                } catch (URISyntaxException e) {
                    LOG.warn("Invalid URI in hardwaretypeinfo entry: "+ee.getHtmlDescription());
                    eHtmlDescription = null;
                }
                ServerGlimpse.AdvertisementInfo.HardwareTypeInfo.HardwareTypeInfoEntry e =
                        new ServerGlimpse.AdvertisementInfo.HardwareTypeInfo.HardwareTypeInfoEntry(
                                ee.getName(), ee.getHrn(),
                                ee.getTextDescription(), eHtmlDescription
                        );
                entries.add(e);
            }
            ServerGlimpse.AdvertisementInfo.HardwareTypeInfo glimpseHardwareTypeInfo =
                    new ServerGlimpse.AdvertisementInfo.HardwareTypeInfo(
                            hardwareTypeInfo.getTextOverview(),
                            htmlOverview,
                            entries
                    );
            glimpse.setHardwareTypeInfo(glimpseHardwareTypeInfo);
        }

        builder.setCountIpv4(-1);
        builder.setCountIpv4Total(-1);
        if (modelRspec.getRoutableAddressesAvailable() != null)
            builder.setCountIpv4( modelRspec.getRoutableAddressesAvailable());
        if (modelRspec.getRoutableAddressesConfigured() != null)
            builder.setCountIpv4Total(modelRspec.getRoutableAddressesConfigured());
        LOG.info("Result count  IPv4 " + builder.getCountIpv4() + " of " + builder.getCountIpv4Total());

        RSpecStatistics stats = builder.create();

        LOG.info("Result countPerHwType hwTypes.keys= " + stats.getHwCountKeys());

        countsToServerGlimpse(stats, glimpse, isAvailableAdvertisement);

        return new Pair<>(stats, glimpse);
    }


    /**
     * This fallback is more resiliant to errors in the RSpec, but it does a much lesser job of counting nodes.
     * Some of the many things it is missing:
     *   - It doesn't fill in most of the ServerGlimpse
     *   - It doesn't categorize the counts
     *   - It cannot check for hardware type to ignore in this fallback
     *   - It doesn't count openglow datapath DPID's
     *   - It doesn't count IP addresses
     *
     * @param rspec
     * @param isAvailableAdvertisement
     * @param targetServer
     * @return
     */
    public Pair<RSpecStatistics, ServerGlimpseBuilder> makeUsingFallback(@Nonnull String rspec,
                                                                         boolean isAvailableAdvertisement,
                                                                         @Nonnull Server targetServer,
                                                                         @Nullable Timestamp timeStamp) {
        BasicStringRspec stringRspec = new BasicStringRspec(rspec);
        if (stringRspec == null)
            throw new RuntimeException("Failed to create stringRspec");
        if (stringRspec.isWellFormed())
            throw new RuntimeException("RSpec is not well formed XML");
        if (stringRspec.isValidRspec())
            throw new RuntimeException("RSpec is not valid XML");
        List<BasicStringRspec.BasicNodeInfo> nodeInfos = stringRspec.getBasicNodeInfo();
        if (nodeInfos == null)
            throw new RuntimeException("Created stringRspec has null nodeInfos list");

        RSpecStatisticsBuilder builder = new RSpecStatisticsBuilder();
        ServerGlimpseBuilder glimpse = new ServerGlimpseBuilder();
        glimpse.setServer(targetServer);
        glimpse.setAdvertisementTimestamp(timeStamp);
        builder.setCount(nodeInfos.size());
        builder.setCountTotal(nodeInfos.size());
        builder.setSkippedCount(0);

        builder.setCountRaw(0);
        builder.setCountRawTotal(0);
        builder.setCountVm(0);
        builder.setCountVmTotal(0);

        //this fallback should be seen as a problem:
        builder.addCountWarning("BasicStringRspec parsing fallback used. Counts will not be correct.");
        builder.addCountError("BasicStringRspec parsing fallback used. Counts will not be correct.");

        RSpecStatistics stats = builder.create();
        countsToServerGlimpse(stats, glimpse, isAvailableAdvertisement);

        return new Pair<>(stats, glimpse);
    }

    /**
     * Create statistics on number of nodes
     * @param rspec A String containing a geni v3 RSpec. (normally advertisement RSpec, use others at own risk.)
     * @return statistics in 2 versions
     */
    public Pair<RSpecStatistics, ServerGlimpseBuilder> make(@Nonnull String rspec,
                                                            boolean isAvailableAdvertisement,
                                                            @Nonnull Server targetServer,
                                                            @Nullable Timestamp timeStamp) {
        try {
            RspecSource rspecSource = new AdvertisementRspecSource(rspec, ModelRspecType.BASIC);
            ModelRspec modelRspec = rspecSource.getModelRspec();
            if (modelRspec == null)
                throw new RuntimeException("Failed to create modelRspec");
            if (modelRspec.getNodes() == null)
                throw new RuntimeException("Created modelRspec has null node list");

            LOG.debug("Advertisement RSpec node count=" + modelRspec.getNodes().size() + "  link count=" + modelRspec.getLinks().size());

            return make(modelRspec, isAvailableAdvertisement, targetServer, timeStamp);
        } catch (Throwable t) {
            ownLOG.error("Error counting in Advertisement RSpec. Falling back to BasicStringRspec method.\n", t);
            LOG.error("\nError counting in Advertisement RSpec (" + t.getMessage() + "). Falling back to BasicStringRspec method.");

            try {
                return makeUsingFallback(rspec, isAvailableAdvertisement, targetServer, timeStamp);
            } catch (Throwable t2) {
                ownLOG.error("Error counting using fallback, in Advertisement RSpec\n", t2);
                LOG.error("\nError counting in Advertisement RSpec (" + t2.getMessage() + "). Both ModelRspec and BasicStringRspec methods failed.");

                RSpecStatisticsBuilder builder = new RSpecStatisticsBuilder();
                ServerGlimpseBuilder glimpse = new ServerGlimpseBuilder();
                glimpse.setServer(targetServer);
                glimpse.setAdvertisementTimestamp(timeStamp);
                builder.setCount(-1);
                builder.setCountTotal(-1);

                //this 2nd fallback should be seen as a problem:
                builder.addCountWarning("Error counting using fallback. No count done.");
                builder.addCountError("Error counting using fallback. No count done.");

                return new Pair<>(builder.create(), glimpse);
            }
        }
    }


    private void countsToServerGlimpse(RSpecStatistics res,
                                       ServerGlimpseBuilder glimpse,
                                       boolean isAvailableAdvertisement) {
        glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.ANY, res.getCountTotal(), res.getCount());
        if (!isAvailableAdvertisement) {
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.RAW_PC, res.getCountRawTotal(), res.getCountRaw());
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.VM, res.getCountVmTotal(), res.getCountVm());

            for (Map.Entry<String, RSpecStatistics.CountPair> e : res.getCountPerHwType().entrySet()) {
                String hwType = e.getKey();
                RSpecStatistics.CountPair p = e.getValue();
                glimpse.setHwCount(hwType, p.getTotal(), p.getAvailable());
            }
        } else {
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.RAW_PC,
                    ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE, res.getCountRaw());
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.VM,
                    ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE, res.getCountVm());

            for (Map.Entry<String, RSpecStatistics.CountPair> e : res.getCountPerHwType().entrySet()) {
                String hwType = e.getKey();
                RSpecStatistics.CountPair p = e.getValue();
                glimpse.setHwCount(hwType, ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE, p.getAvailable());
            }
        }
        if (res.getCountIpv4() >= 0 || res.getCountIpv4Total() >= 0) {
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.IPV4,
                    ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE, res.getCountIpv4());

            if (res.getCountIpv4Total() >= 0) {
                glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.IPV4,
                        ServerGlimpse.AdvertisementInfo.CountScope.ALL, res.getCountIpv4Total());
            }
        }

        if (res.getCountOpenflow() >= 0 || res.getCountOpenflowTotal() >= 0) {
            glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.OPENFLOW,
                    ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE, res.getCountOpenflow());

            if (res.getCountOpenflowTotal() >= 0) {
                glimpse.setCount(ServerGlimpse.AdvertisementInfo.CountedObject.OPENFLOW,
                        ServerGlimpse.AdvertisementInfo.CountScope.ALL, res.getCountOpenflowTotal());
            }
        }
    }

    /**
     * Must a node with the given hwTypeName be ignored completely (=not counted at all)?
     * @param n
     * @param hwTypeName
     * @param targetServer
     * @return
     */
    private boolean mustIgnoreNode(RspecNode n, String hwTypeName, Server targetServer) {
        if (hwTypeName != null &&
                (hwTypeName.equalsIgnoreCase("switch") ||
                        hwTypeName.equalsIgnoreCase("SWITCHPORT") ||
                        hwTypeName.equalsIgnoreCase("ipv4") || //FEDIBBTDEV-1586
                        hwTypeName.equalsIgnoreCase("orca-iscsi-luns") ||
                        hwTypeName.equalsIgnoreCase("orca-transit-net-vlan") ||
                        hwTypeName.equalsIgnoreCase("xcvr2450") || hwTypeName.equalsIgnoreCase("flex900") || hwTypeName.equalsIgnoreCase("80211") || //FEDIBBTDEV-2258: virtual wall "airswitch"
                        hwTypeName.toLowerCase().startsWith("interconnect"))) {
            return true;
        } else {
            return false;
        }

    }

    public boolean mustExcludeHardwareTypeFromServerGlimpse(RspecNode n, String hardwareType, Server server) {
        if (hardwareType == null) return true;

        String hardwareTypeToCheck = hardwareType.trim();
        /** emulab adds a lot of useless hardware types. We exclude them all here. */
        if (server.getServerType() != null &&
                (server.getServerType().trim().equalsIgnoreCase("emulab") ||
                        server.getServerType().trim().equalsIgnoreCase("instageni"))
        ) {
            if (hardwareTypeToCheck.equals("pc")) return true;
            if (hardwareTypeToCheck.equals("vm")) return true;
            if (hardwareTypeToCheck.equals("pcvm")) return true;
            if (hardwareTypeToCheck.equals("delay")) return true;
            if (hardwareTypeToCheck.equals("lan")) return true;
            if (hardwareTypeToCheck.startsWith("pcvm")) return true;
            if (hardwareTypeToCheck.startsWith("delay-")) return true;
            if (hardwareTypeToCheck.startsWith("bridge-")) return true;
            if (hardwareTypeToCheck.startsWith("bridge")) return true;
            if (hardwareTypeToCheck.endsWith("-vm")) return true;
        }

        return false;
    }

    public boolean mustCountVmTypeSlots(RspecNode n, String hwTypeName, Server targetServer) {
        if (hwTypeName.equalsIgnoreCase("orca-vm-cloud")) {
            //exogeni specific
            return true;
        }
        if (hwTypeName.equalsIgnoreCase("pcvm")) {
            //emulab specific
            return true;
        }
        if (hwTypeName.equalsIgnoreCase("vm") && targetServer.getUrnTld().equalsIgnoreCase("ocf:i2cat:vtam")) {
            //i2cat VTAM specific
            return true;
        }
        if (hwTypeName.startsWith("vmhost")) {
            //Iris TCD specific
            return true;
        }
        if (hwTypeName.equals("docker_cluster")) {
            //Docker AM specific
            return true;
        }
        return false;
    }

    public int getTotalVmCountPerNode(int curCountVm, RspecNode n, String hwTypeName, Server targetServer) {
        //hardcoded HACK! (needs to be advertised so we can automatically find this)
        if (targetServer.getUrnTld().equals("wall2.ilabt.iminds.be") || targetServer.getUrnTld().equals("wall1.ilabt.iminds.be"))
            return 40; //the virtual walls have 40 configured
        if (targetServer.isServerType("instageni"))
            return 100; //instageni has 100
        if (hwTypeName.equalsIgnoreCase("orca-vm-cloud"))
            return 52; //guess for exogeni
        if (hwTypeName.equalsIgnoreCase("vm") && targetServer.getUrnTld().equalsIgnoreCase("ocf:i2cat:vtam"))
            return 40; //guess for i2cat VTAM
        if (hwTypeName.equalsIgnoreCase("vmhost-type1-vm"))
            return 70; //specified by Diarmuid for Iris TCD
        if (hwTypeName.equalsIgnoreCase("docker_cluster"))
            return 20; //Docker AM hardcoded (todo needs to be set per server instead)
        return curCountVm;
    }

    public boolean includeVmCountInTotal(RspecNode n, String hwTypeName, Server targetServer) {
        //don't count VM's on emulab
        if (targetServer.isServerType("emulab"))
            return false;
        if (targetServer.isServerType("instageni"))
            return false;
//            if (hwTypeName.equalsIgnoreCase("orca-vm-cloud"))
//                return false;
        return true;
    }

    public boolean isVm(RspecNode n, String hwTypeName, Server targetServer) {
        if (Objects.equals(n.getExclusive(), Boolean.FALSE))
            return true;
        if (hwTypeName.equalsIgnoreCase("orca-vm-cloud"))
            return true;
        return false;
    }

    private String friendlyName(RspecNode node) {
        if (node.getComponentName() != null)
            return node.getComponentName();
        if (node.getComponentId() != null) {
            return node.getComponentId().getResourceName();
        }
        return null;
    }
}
